from io import BytesIO
from pathlib import Path
from typing import Literal

import markdown2
from docling.datamodel.document import DocumentStream

from docutranslate.Agents import Agent, AgentArgs
from docutranslate.Agents import MDRefineAgent, MDTranslateAgent
from docutranslate.utils.convert import file2markdown_embed_images
from docutranslate.utils.markdown_splitter import split_markdown_text
from docutranslate.utils.markdown_utils import uris2placeholder, placeholder2_uris, MaskDict


# import mdformat


class FileTranslater:
    def __init__(self, file_path: Path | str | None = None, chunksize: int = 3500, base_url="", key=None,
                 model_id="", temperature=0.7, max_concurrent=10, docling_artifact: Path | str | None = None,
                 timeout=2000, tips=True):
        if isinstance(file_path, str):
            file_path = Path(file_path)
        self.file_path: Path = file_path
        self.file_path: Path = file_path
        self._mask_dict = MaskDict()
        self.markdown: str = ""
        self.chunksize = chunksize
        self.max_concurrent = max_concurrent
        self.base_url: str = base_url
        self.key: str = key if key is not None else "xx"
        self.model_id: str = model_id
        self.temperature = temperature
        self.docling_artifact = docling_artifact
        self.timeout = timeout
        if tips:
            print("""
=======
[docutranslate](https://github.com/xunbu/docutranslate)
>以下操作会自动从[huggingface](https://huggingface.co)下载模型，windows需要使用**管理员模式**打开IDE运行脚本，并按需换源
- 第一次使用该库读取、翻译非markdown文本
- 第一次使用该库的公式识别或代码识别功能
=======
""")

    def _markdown_format(self):
        # 该方法还需要改进
        # self.markdown=mdformat.text(self.markdown)
        pass

    def _mask_uris_in_markdown(self):
        self.markdown = uris2placeholder(self.markdown, self._mask_dict)
        return self

    def _unmask_uris_in_markdown(self):
        self.markdown = placeholder2_uris(self.markdown, self._mask_dict)
        return self

    def _split_markdown_into_chunks(self) -> list[str]:
        chunks: list[str] = split_markdown_text(self.markdown, self.chunksize)
        print(f"markdown分为{len(chunks)}块")
        return chunks

    def default_agent_params(self) -> AgentArgs:
        result: AgentArgs = {
            "baseurl": self.base_url,
            "key": self.key,
            "model_id": self.model_id,
            "temperature": self.temperature,
            "max_concurrent": self.max_concurrent,
            "timeout": self.timeout
        }
        return result

    def read(self, name: str, file: bytes, formula=False, code=False, save=False,
             save_format: Literal["markdown", "html"] = "markdown", refine=False,
             refine_agent: Agent | None = None):
        ds = DocumentStream(name=name, stream=BytesIO(file))
        file_path = Path(name)
        # 如果是markdown，直接读取
        if file_path.suffix == ".md":
            self.markdown = file.decode()
        else:
            print(f"正在将{file_path.resolve().name}转换为markdown")
            self.markdown = file2markdown_embed_images(ds, formula, code, artifacts_path=self.docling_artifact)
            print("已转换为markdown")
        if refine:
            self.refine_markdown_by_agent(refine_agent)
        if save:
            if save_format == "html":
                self.save_as_html(filename=f"{file_path.stem}.html")
            else:
                self.save_as_markdown(filename=f"{file_path.stem}.md")
        return self

    def read_file(self, file_path: Path | str | None = None, formula=False, code=False, save=False,
                  save_format: Literal["markdown", "html"] = "markdown", refine=False,
                  refine_agent: Agent | None = None):
        if file_path is None:
            if self.file_path is None:
                raise Exception("未设置文件路径")
            file_path = self.file_path
        if isinstance(file_path, str):
            file_path = Path(file_path)
        print(f"读取文件：{file_path.name}")
        # 如果是markdown，直接读取
        if file_path.suffix == ".md":
            with open(file_path, "r") as f:
                self.markdown = f.read()
        else:
            print(f"正在将{file_path.resolve().name}转换为markdown")
            self.markdown = file2markdown_embed_images(file_path, formula, code, artifacts_path=self.docling_artifact)
            print("已转换为markdown")
        if refine:
            self.refine_markdown_by_agent(refine_agent)
        if save:
            if save_format == "html":
                self.save_as_html(filename=f"{file_path.stem}.html")
            else:
                self.save_as_markdown(filename=f"{file_path.stem}.md")
        return self

    def refine_markdown_by_agent(self, refine_agent: Agent | None = None) -> str:
        print("正在修正markdown")
        self._mask_uris_in_markdown()
        chuncks = self._split_markdown_into_chunks()
        if refine_agent is None:
            refine_agent = MDRefineAgent(**self.default_agent_params())
        result: list[str] = refine_agent.send_prompts(chuncks)
        self.markdown = "\n".join(result)
        self._unmask_uris_in_markdown()
        print("markdown已修正")
        return self.markdown

    def translate_markdown_by_agent(self, translate_agent: Agent | None = None, to_lang="中文"):
        print("正在翻译markdown")
        self._mask_uris_in_markdown()
        chuncks = self._split_markdown_into_chunks()
        if translate_agent is None:
            translate_agent = MDTranslateAgent(to_lang=to_lang, **self.default_agent_params())
        result: list[str] = translate_agent.send_prompts(chuncks)
        self.markdown = "\n".join(result)
        self._unmask_uris_in_markdown()
        print("翻译完成")
        return self.markdown

    def save_as_markdown(self, filename: str | Path | None = None, output_dir: str | Path = "./output"):
        if isinstance(filename, str):
            filename = Path(filename)
        if isinstance(output_dir, str):
            output_dir = Path(output_dir)
        if filename is None:
            if self.file_path is not None:
                filename = self.file_path.name
            else:
                filename = "output.md"
        # 确保输出目录存在
        output_dir.mkdir(parents=True, exist_ok=True)
        full_name = output_dir / filename
        # 输出前格式化markdown
        self._markdown_format()
        with open(full_name, "w") as file:
            file.write(self.markdown)
        print(f"文件已写入{full_name.resolve()}")
        return self

    def export_to_markdown(self):
        # 输出前格式化markdown
        self._markdown_format()
        return self.markdown

    def save_as_html(self, filename: str | Path | None = None, output_dir: str | Path = "./output"):
        if isinstance(filename, str):
            filename = Path(filename)
        if isinstance(output_dir, str):
            output_dir = Path(output_dir)
        if filename is None:
            if self.file_path is not None:
                filename = self.file_path.name
            else:
                filename = "output.html"
        # 确保输出目录存在
        output_dir.mkdir(parents=True, exist_ok=True)
        full_name = output_dir / filename
        html = self.export_to_html(str(filename.resolve().stem))
        with open(full_name, "w") as file:
            file.write(html)
        print(f"文件已写入{full_name.resolve()}")
        return self

    def export_to_html(self, title="title") -> str:
        markdowner = markdown2.Markdown(extras=['tables', 'fenced-code-blocks', 'mermaid', "code-friendly"])

        html = f"""<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <title>{title}</title>
    <link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/@picocss/pico@latest/css/pico.min.css">
        <style>
        html {{
            padding:2vh 10vw;
            font-size: 15px;
        }}
    </style>
    <script type="text/x-mathjax-config">
      MathJax.Hub.Config({{
        tex2jax: {{
          inlineMath: [ ['$','$'], ["\\\\(","\\\\)"] ],
          processEscapes: true
        }}
      }});
    </script>
        
    <script type="text/javascript"
            src="https://cdn.mathjax.org/mathjax/latest/MathJax.js?config=TeX-AMS-MML_HTMLorMML">
    </script>
</head>
<body>
{markdowner.convert(self.markdown)}
</body>
<script type="module" defer>
  import mermaid from 'https://cdn.jsdelivr.net/npm/mermaid@9/dist/mermaid.esm.min.mjs';
  mermaid.initialize({{
    securityLevel: 'loose',
    startOnLoad: true
  }});
  let observer = new MutationObserver(mutations => {{
    for(let mutation of mutations) {{
      mutation.target.style.visibility = "visible";
    }}
  }});
  document.querySelectorAll("pre.mermaid-pre div.mermaid").forEach(item => {{
    observer.observe(item, {{ 
      attributes: true, 
      attributeFilter: ['data-processed'] 
    }});
  }});
</script>

</html>
"""
        return html

    def translate_file(self, file_path: Path | str | None = None, to_lang="中文", output_dir="./output",
                       formula=False,
                       code=False, output_format: Literal["markdown", "html"] = "markdown", refine=False,
                       refine_agent: Agent | None = None, translate_agent: Agent | None = None):
        if file_path is None:
            assert self.file_path is not None, "未输入文件路径"
            file_path = self.file_path
        if isinstance(file_path, str):
            file_path = Path(file_path)
        self.read_file(file_path, formula=formula, code=code)
        if refine:
            self.refine_markdown_by_agent(refine_agent)
        self.translate_markdown_by_agent(translate_agent, to_lang=to_lang)
        if output_format == "markdown":
            filename = f"{file_path.stem}_{to_lang}.md"
            self.save_as_markdown(filename=filename, output_dir=output_dir)
        elif output_format == "html":
            filename = f"{file_path.stem}_{to_lang}.html"
            self.save_as_html(filename=filename, output_dir=output_dir)
        return self



from docutranslate.utils.markdown_splitter import split_markdown_text
from docutranslate import FileTranslater

ts=FileTranslater()
ts.read_file(r".\resource\regex.md")
ts._mask_uris_in_markdown()
# ts._unmask_uris_in_markdown()

a=split_markdown_text(ts.markdown,max_block_size=2000)
print("\n==================\n".join(a))


import re
import threading
import uuid


class MaskDict:
    def __init__(self):
        self._dict = {}
        self._lock = threading.Lock()

    def create_id(self):
        with self._lock:
            while True:
                id = uuid.uuid1().hex[:6]
                if id not in self._dict:
                    return id

    def get(self, key):
        with self._lock:
            return self._dict.get(key)

    def set(self, key, value):
        with self._lock:
            self._dict[key] = value

    def delete(self, key):
        with self._lock:
            if key in self._dict:
                del self._dict[key]

    def __contains__(self, item):
        with self._lock:
            return item in self._dict


# def uris2placeholder(markdown:str, mask_dict:MaskDict):
##替换整个uri
#     def uri2placeholder(match: re.Match):
#         id = mask_dict.create_id()
#         mask_dict.set(id, match.group())
#         return f"<ph-{id}>"
#
#     uri_pattern = r'!?\[.*?\]\(.*?\)'
#     markdown = re.sub(uri_pattern, uri2placeholder, markdown)
#     return markdown

def uris2placeholder(markdown: str, mask_dict: MaskDict):
    ##只替换uri里的链接部分，保留标题
    def uri2placeholder(match: re.Match):
        id = mask_dict.create_id()
        mask_dict.set(id, match.group(2))
        return f"{match.group(1)}(<ph-{id}>)"

    uri_pattern = r'(!?\[.*?\])\((.*?)\)'
    markdown = re.sub(uri_pattern, uri2placeholder, markdown)
    return markdown


def placeholder2_uris(markdown: str, mask_dict: MaskDict):
    def placeholder2uri(match: re.Match):
        id = match.group(1)
        uri = mask_dict.get(id)
        if uri is None:
            return match.group()
        return uri

    ph_pattern = r"<ph-([a-zA-Z0-9]+)>"
    markdown = re.sub(ph_pattern, placeholder2uri, markdown)
    return markdown


if __name__ == '__main__':
    pass



a=r"""
# 角色
你是一个修正markdown文本的专家。
# 工作
找到markdown片段的不合理之处。
对于缺失、中断的句子，应该查看缺失的语句是否可能被错误的放在了其他位置，并通过句子拼接修复不合理之处。
去掉异常字词，修复错误格式。
# 要求
If refine is unnecessary, return the original text.
NO explanations. NO notes.
忠实于原文。
不要修改标题的级别（如一级标题不要修改为二级标题）
形如<ph-abc123>的占位符不要改变。
code、latex和HTML保持结构。
# 输出
修正后的markdown纯文本（不是markdown代码块）
# 示例
## 修正文本流
输入：
什么名字
你叫
输出：
你叫什么名字
## 去掉异常字词(保持占位符和latex符号)
输入：
一道\题@#目<ph-12asd2>:\(x_1+1=2\)
输出:
一道题目<ph-12asd2>:：\(x_1+1=2\)
\no_think"""
print(len(a.encode()))


from typing import Unpack

from .agent import Agent, AgentArgs


class MDRefineAgent(Agent):
    def __init__(self,**kwargs:Unpack[AgentArgs]):
        super().__init__(**kwargs)
        self.system_prompt=r"""
# 角色
你是一个修正markdown文本的专家。
# 工作
找到markdown片段的不合理之处。
对于缺失、中断的句子，应该查看缺失的语句是否可能被错误的放在了其他位置，并通过句子拼接修复不合理之处。
去掉异常字词，修复错误格式。
# 要求
If refine is unnecessary, return the original text.
NO explanations. NO notes.
不要修改标题的级别（如一级标题不要修改为二级标题）
形如<ph-abc123>的占位符不要改变。
code、latex和HTML保持结构。
# 输出
修正后的markdown纯文本（不是markdown代码块）
# 示例
## 修正文本流
输入：
什么名字
你叫
输出：
你叫什么名字
## 去掉异常字词(保持占位符和latex符号)
输入：
一道\题@#目<ph-12asd2>:\(x_1+1=2\)
输出:
一道题目<ph-12asd2>:：\(x_1+1=2\)
\no_think"""


class MDTranslateAgent(Agent):
    def __init__(self,to_lang="中文",**kwargs:Unpack[AgentArgs]):
        super().__init__(**kwargs)
        self.system_prompt=f"""
# 角色
You are a professional, authentic machine translation engine.
# 工作
翻译输入的markdown文本
目标语言{to_lang}
# 要求
If translation is unnecessary (e.g. proper nouns, codes, etc.), return the original text. 
NO explanations. NO notes.
不要修改标题的级别（如一级标题不要修改为二级标题）
形如<ph-abc123>的占位符不要改变。
code、latex和HTML只翻译说明文字，其余保持原文
# 输出
翻译后的markdown纯文本（不是markdown代码块）
# 示例
## 英文翻译为中文：
输入：
hello<ph-aaaaaa>, what's your name?
输出：
你好<ph-aaaaaa>，你叫什么名字？
\\no_think"""


from docutranslate import FileTranslater


translater = FileTranslater(base_url="https://open.bigmodel.cn/api/paas/v4",
                            key="969ba51b61914cc2b710d1393dca1a3c.hSuATex5IoNVZNGu",
                            model_id="glm-4-flash",
                            # model_id="glm-4-air",
                            chunksize=3500,
                            max_concurrent=20
                            # artifact_path=r"C:\Users\jxgm\.cache\docling\models"
                            )
# translater = FileTranslater(base_url="https://dashscope.aliyuncs.com/compatible-mode/v1",
#                             key="sk-a3dd6bdedb5f446cbe678aedfab32038",
#                             # model_id="qwen-turbo",
#                             model_id="qwen-plus",
#                             max_concurrent=20)
# translater = FileTranslater(base_url="http://127.0.0.1:1234/v1",
#                             key="969ba51b61914cc2b710d1393dca1a3c.hSuATex5IoNVZNGu",
#                             model_id="qwen3-1.7b")
# translater=FileTranslater(base_url="https://api.deepseek.com/v1",
#                key="sk-d809e448131d4fbc903a0ed476964294",
#                model_id="deepseek-chat")
#
translater.translate_file(r".\resource\test6.md",
                          formula=True,
                          code=True,
                          refine=True,
                          to_lang="中文",
                          output_format="markdown")

# translater.read_file(r".\resource\test.pdf",formula=True,code=True,save=True,refine=False,save_format="markdown")

# with open(r"C:\Users\jxgm\Desktop\FileTranslate\tests\resource\test2.pdf","rb") as f:
#     translater.read(name="test2.pdf",file=f.read(),formula=True,code=True, save=True, refine=False, save_format="html")


# import mistletoe
# from mistletoe.latex_renderer import LaTeXRenderer
#
# with open(r"C:\Users\jxgm\Desktop\FileTranslate\tests\resource\互联网认证授权机制.md","r",encoding="utf-8") as f:
#     md=f.read()
# html = mistletoe.markdown(md)
# # with open(r"./output.html","w",encoding="utf-8") as f:
# #     f.write(html)
# from fpdf import FPDF
# pdf = FPDF()
# pdf.add_page()
# pdf.write_html(html)
# pdf.output("pdf-from-markdown-with-mistletoe.pdf")


import os
from huggingface_hub.errors import LocalEntryNotFoundError
from docling.datamodel.base_models import InputFormat
from docling.datamodel.pipeline_options import PdfPipelineOptions
from docling_core.types.doc import ImageRefMode
from pathlib import Path
from docling.document_converter import DocumentConverter, PdfFormatOption
from docling.datamodel.document import DocumentStream

IMAGE_RESOLUTION_SCALE = 4


def file2markdown_embed_images(file_path: Path | str|DocumentStream, formula=False, code=False,artifacts_path:Path|str|None=None) -> str:
    pipeline_options = PdfPipelineOptions(artifacts_path=artifacts_path)
    # pipeline_options.do_ocr=False
    pipeline_options.images_scale = IMAGE_RESOLUTION_SCALE
    pipeline_options.generate_picture_images = True
    if formula:
        pipeline_options.do_formula_enrichment=True
    if code:
        pipeline_options.do_code_enrichment=True
    converter = DocumentConverter(format_options={
        InputFormat.PDF: PdfFormatOption(pipeline_options=pipeline_options)
    })
    try:
        result = converter.convert(file_path).document.export_to_markdown(image_mode=ImageRefMode.EMBEDDED)
    except LocalEntryNotFoundError:
        print(f"无法连接huggingface，正在尝试换源")
        os.environ['HF_ENDPOINT'] = 'https://hf-mirror.com'
        result = converter.convert(file_path).document.export_to_markdown(image_mode=ImageRefMode.EMBEDDED)
    return result
